/*
** iplog_scan.c - iplog scan/flood detector.
** Copyright (C) 1999 Odin <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
*/

#include <config.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <syslog.h>
#include <pthread.h>
#include <rwlock.h>

#include <iplog.h>
#include <iplog_scan.h>
#include <iplog_options.h>

#define ENDSCAN \
	"%s mode expired for %s - received a total of %lu packets (%lu bytes)."

extern u_long flags;

__inline__ static bool no_res(void) {
	return (0);
}

static u_int scan_entries;
static struct scan_data *scan_table[SCAN_TSIZE];

static const struct scan_info scan[] = {
	{ "TCP: port scan",			tcp_res,	PS_TIMEOUT,		PS_THOLD	},
	{ "TCP: null scan",			tcp_res,	NS_TIMEOUT,		NS_THOLD	},
	{ "TCP: FIN scan",			tcp_res,	FS_TIMEOUT,		FS_THOLD	},
	{ "TCP: Xmas scan",			tcp_res,	XS_TIMEOUT,		XS_THOLD	},
	{ "UDP: scan/flood",		udp_res,	UDP_TIMEOUT,	UDP_THOLD	},
	{ "ICMP/UDP: smurf attack",	no_res,		SMURF_TIMEOUT,	SMURF_THOLD	},
	{ "ICMP: ping flood",		icmp_res,	PING_TIMEOUT,	PING_THOLD	}
};

pthread_mutex_t scan_lock = PTHREAD_MUTEX_INITIALIZER;

bool check_scan(u_char type, u_long s_addr, u_long len) {
	u_long hash, addr;
	struct scan_data *cur;
	struct scan_t *data;
	bool ret = false, found = false, log_scan = false;

	if (type == SCAN_SMURF)
		addr = s_addr & 0xffffff;
	else
		addr = s_addr;

	hash = HASH(addr, SCAN_TSIZE);

	pthread_mutex_lock(&scan_lock);
	for (cur = scan_table[hash] ; cur != NULL ; cur = cur->next) {
		if (cur->addr == addr) {
			found = true;
			data = &cur->type[type];

			pthread_mutex_unlock(&scan_lock);
			rwl_readlock(&cur->lock);
			pthread_mutex_lock(&data->lock);

			data->bytes += len;
			data->expire = time(NULL) + scan[type].timeout;

			if (++data->count >= scan[type].threshold) {
				ret = true;
				if (data->logged == false) {
					data->logged = true;
					log_scan = true;
				}
			}

			pthread_mutex_unlock(&data->lock);
			rwl_readunlock(&cur->lock);

			break;
		}
	}

	if (found == false) {
		if (scan_entries < SCAN_MAXENT) {
			struct scan_data temp, *t_cur;

			memset(&temp, 0, sizeof(temp));
			temp.addr = addr;
			t_cur =
				list_prepend(&temp, (void **) &scan_table[hash], sizeof(temp));
			++scan_entries;
			data = &t_cur->type[type];

			pthread_mutex_unlock(&scan_lock);
			rwl_readlock(&t_cur->lock);
			pthread_mutex_lock(&data->lock);

			data->count = 1;
			data->bytes = len;
			data->expire = time(NULL) + scan[type].timeout;

			pthread_mutex_unlock(&data->lock);
			rwl_readunlock(&t_cur->lock);
		} else
			pthread_mutex_unlock(&scan_lock);
	}

	if (log_scan == true) {
		u_char buf[256];
		struct in_addr in;

		in.s_addr = addr;

		mysyslog(PRIORITY, "%s detected from %s",
			scan[type].name,
			host_lookup(&in, scan[type].resolv(), buf, sizeof(buf)));

		return (true);
	}

	return (ret);
}

void expire_scans(void) {
	size_t i, j;
	struct scan_data *cur;
	time_t now;
	bool remove;
	u_char buf[256];
	struct in_addr in;

	pthread_mutex_lock(&scan_lock);
	time(&now);

	for (i = 0 ; i < SCAN_TSIZE ; i++) {
		for (cur = scan_table[i] ; cur != NULL ; cur = cur->next) 
			rwl_writelock(&cur->lock);
	}

	for (i = 0 ; i < SCAN_TSIZE ; i++) {
		cur = scan_table[i];

		for (remove = true ; cur != NULL ;) {
			for (j = 0 ; j < SCAN_TOTAL ; j++) {
				if (cur->type[j].count != 0 && now >= cur->type[j].expire) {
					if (cur->type[j].count > scan[j].threshold) {
						in.s_addr = cur->addr;
						mysyslog(LOG_INFO, ENDSCAN, scan[j].name,
							host_lookup(&in, scan[j].resolv(), buf,sizeof(buf)),
							cur->type[j].count, cur->type[j].bytes);
					}
					cur->type[j].expire = 0;
					cur->type[j].logged = 0;
					cur->type[j].count = 0;
					cur->type[j].bytes = 0;
				} else if (cur->type[j].count != 0)
					remove = false;
			}

			if (remove == true) {
				cur = list_delete(cur, (void **) &scan_table[i]);
				--scan_entries;
			} else {
				rwl_writeunlock(&cur->lock);
				cur = cur->next;
			}
		}
	}

	pthread_mutex_unlock(&scan_lock);
}

#ifdef HAVE_PTHREAD_CANCEL
void destroy_scan_table(void) {
	size_t i;
	struct scan_data *cur;

	pthread_mutex_lock(&scan_lock);
	for (i = 0 ; i < SCAN_TSIZE ; i++) {
		for (cur = scan_table[i] ; cur != NULL ; cur = cur->next)
			rwl_writelock(&cur->lock);
		list_destroy(scan_table[i], NULL);
		scan_table[i] = NULL;
	}
	scan_entries = 0;
	pthread_mutex_unlock(&scan_lock);
}
#endif

/* vim:ts=4:sw=8:tw=0 */
